import random

# Por simplicidad colocaremos todas las clases en un único módulo, aunque es
# recomendable utilizar un módulo para cada clase


class Dron:
    """ Clase que representa a un dron """

    def __init__(self, idD, carga):
        self.idD = idD
        self.carga = carga
        self.estado = 'libre'
        self.pedido = None

    def __str__(self):
        """ No se pide pero es útil para comprobar que el código es correcto"""
        return ('dron ' + self.idD + ' con carga ' + str(self.carga) +
                ' estado ' + self.estado + ' y orden ' + str(self.pedido))

    # Creamos propiedades para carga y estado, podríamos también comprobar los
    # tipos para el resto de atributos, pero con esto bastaba para el examen
    @property
    def carga(self):
        return self.__carga

    @carga.setter
    def carga(self, v):
        if v >= 0 and v <= 100:
            self.__carga = v
        else:
            self.__carga = 0

    @property
    def estado(self):
        return self.__estado

    @estado.setter
    def estado(self, v):
        if v.lower() == 'libre' or v.lower() == 'ocupado':
            self.__estado = v
        else:
            self.__estado = 'libre'


class Pedido:
    """ Clase que representa un pedido Pedido """

    def __init__(self, idO, p, direccion):
        self.idO = idO
        self.direccion = direccion
        self.prioridad = p

    def __str__(self):
        """ No se pide pero es útil para comprobar que el código es correcto"""
        return ('Pedido ' + self.idO + ' con prioridad ' + str(self.prioridad)
                + ' y dirección ' + self.direccion)


class Mapa:
    """ Clase que representa el mapa de un vecindario """

    def __init__(self, whId, rows, columns, houses):
        # El único atributo será el mapa del vecindario. Podría haber también
        # atributos para columnas, filas, etc. pero como se pueden calcular
        # usando el mapa, no los creamos para evitar repetir información
        self.mapa = self.crearMapaArea(whId, rows, columns, houses)

    def crearMapaArea(self, whId, rows, columns, houses):
        """ Crea el mapa del vecindario, añadiendo el almacén y las casas en
            posiciones aleatorias """
        mapa = []
        # Rellenamos todo con calles
        for _ in range(rows):
            row = []
            for _ in range(columns):
                row.append('s')
            mapa.append(row)
        # Ahora ponemos el almacén
        mapa[random.randrange(rows)][random.randrange(columns)] = whId
        # Y por último las casas
        count = 1
        while count <= houses:
            r = random.randrange(rows)
            c = random.randrange(columns)
            if 's' in mapa[r][c]:
                mapa[r][c] = 'h' + str(count)
                count += 1
        return mapa

    def printMapa(self):
        """ No se pide pero es útil para comprobar que el código es correcto"""
        for i in self.mapa:
            print(i)

    def calcularDistancia(self, warehouse, houseId):
        """ Calcula la distancia entre el almacen y una casa """
        i, foundW, foundH = 0, False, False
        while i < len(self.mapa) and (not foundW or not foundH):
            j = 0
            while j < len(self.mapa[i]) and (not foundW or not foundH):
                if self.mapa[i][j] == warehouse and not foundW:
                    w = (i, j)
                    foundW = True
                elif self.mapa[i][j] == houseId and not foundH:
                    h = (i, j)
                    foundH = True
                j += 1
            i += 1
        d = ((w[0] - h[0]) ** 2 + (w[1] - h[1]) ** 2) ** 0.5
        return d


class Almacen:
    """Clase que representa el almacén, la mayoría de los métodos están aquí"""

    def __init__(self, name, drons, nRows, nColumns, nHouses):
        self.nombre = name
        self.drones = self.crearDrones(drons)
        self.pedidos = []
        self.mapa = Mapa(name, nRows, nColumns, nHouses)

    def crearDrones(self, n):
        """ Crea una lista de objetos dron """
        drones = []
        for i in range(0, n):
            drones.append(Dron('dron' + str(i + 1), random.randint(0, 100)))
        return drones

    def insertarPedido(self, pedido):
        """ Inserta un Pedido en la lista de Pedidos """
        count, found = 0, False
        while not found and count < len(self.pedidos):
            if self.pedidos[count].prioridad < pedido.prioridad:
                found = True
            else:
                count += 1
        self.pedidos.insert(count, pedido)

    def elegirDron(self, direccion):
        """ Elige un dron para procesar un Pedido """
        d, maxValue, indexMaxValue, found, index = 0, 0, -1, False, None
        distance = self.mapa.calcularDistancia(self.nombre, direccion) * 2
        while not found and d < len(self.drones):
            # Solo consideramos drones libres
            if self.drones[d].estado == 'libre':
                if self.drones[d].carga >= distance:
                    found = True
                    index = d
                # Aunque no tenga carga para servir el pedido calculamos si es
                # el de mayor carga
                elif maxValue < self.drones[d].carga:
                    maxValue = self.drones[d].carga
                    indexMaxValue = d
            d += 1
        # Si ninguno tiene carga suficiente, recargamos el de mayor carga
        if not found:
            if indexMaxValue != - 1:
                self.drones[indexMaxValue].carga = 100
                index = indexMaxValue
        # Si no hay dron disponible devolveremos None y distancia
        return index, distance

    def procesarPedido(self, pedido):
        """ Procesa el Pedido """
        self.insertarPedido(pedido)
        index, distance = self.elegirDron(pedido.direccion)
        if index is not None:
            self.drones[index].pedido = pedido
            self.drones[index].carga -= distance
            self.drones[index].estado = 'ocupado'
            print(pedido, 'procesado por dron', self.drones[index].idD)
        else:
            print('Pedido %s no procesado: todos los drones están ocupados'
                  % pedido.idO)


# Programa principal

name = input('¿Nombre del almacén? ')
drons = int(input('¿Drones? '))
rowMap = int(input('¿Filas? '))
colMap = int(input('¿Columnas? '))
houses = int(input('¿Casas? '))

# Creamos el almacén
w = Almacen(name, drons, rowMap, colMap, houses)
# Creamos los pedidos
for i in range(5):
    direccion = 'h' + str(random.randint(1, houses))
    w.procesarPedido(Pedido('id' + str(i), random.randint(1, 10), direccion))
